/*
 *
 * Copyright (C) 2004 Alexander Yip (yipal@lcs.mit.edu)
 *
 *  Permission is hereby granted, free of charge, to any person obtaining
 *  a copy of this software and associated documentation files (the
 *  "Software"), to deal in the Software without restriction, including
 *  without limitation the rights to use, copy, modify, merge, publish,
 *  distribute, sublicense, and/or sell copies of the Software, and to
 *  permit persons to whom the Software is furnished to do so, subject to
 *  the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 *  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 *  OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 *  WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 */

#include <async.h>
#include "controlclient.h"
#include <blacklist.h>
#include <location.h>

controlclient::controlclient(ptr<chord> _chordnode, str name, int port) 
  : nexthost(NULL), myname(name), myport(port), chordnode(_chordnode),
    busy(false) {
  unsigned r = (unsigned) (  ( (float)(random() & 0xffffffff) / 0xffffffff) * hostlist::only().size());

  advance_next();
  for(unsigned i=0; i<r; i++)
    advance_next();

  delaycb(GOSSIP_DELAY, wrap(this, &controlclient::sync_lists));
}

controlclient::~controlclient() {

}

void
controlclient::sync_lists() {
  if (busy) {
    delaycb(GOSSIP_DELAY, wrap(this, &controlclient::sync_lists));
    return;
  }
  busy = true;

  // get next up node
  advance_next();
  warn << "Trying to sync with " << nexthost->name << ":" << nexthost->port << "\n";
  tcpconnect (nexthost->name, nexthost->port + 3, wrap(this, &controlclient::connect_cb));

  delaycb(GOSSIP_DELAY, wrap(this, &controlclient::sync_lists));
}

void
controlclient::connect_cb(int fd) {
  int i = 0;
  if (fd < 0) {
    warn << "bad fd during sync_lists\n";
    busy = false;
    return;
  }

  ptr<control_sync_arg> arg = New refcounted<control_sync_arg>();
  arg->caller_name = myname;
  arg->caller_port = myport;

  arg->hostlistver = hostlist::only().get_version();

  hostentry *p = hostlist::only().first();
  arg->node_ips.setsize(hostlist::only().size());
  arg->node_ports.setsize(hostlist::only().size());
  arg->versions.setsize(hostlist::only().size());

  while(p) {
    arg->node_ips[i] = p->name;
    arg->node_ports[i] = p->port;
    arg->versions[i] = blacklist::only().getversion(p->name, p->port);

    //warn << "sync args: " << arg->hostlistver << " " << arg->versions[i] 
    //	 << " " << p->name << ":" << p->port << "\n";

    i++;
    p = hostlist::only().next(p);
  }

  ptr<control_sync_res> res = New refcounted<control_sync_res>(CONTROL_OK);

  ptr<aclnt> c = aclnt::alloc (axprt_stream::alloc(fd), control_program_1);
  c->call (CONTROLPROC_SYNC, arg, res, wrap(this, &controlclient::sync_lists_cb, arg, res, c));

}

void
controlclient::sync_lists_cb(ptr<control_sync_arg> arg, ptr<control_sync_res> res, 
			 ptr<aclnt> clnt, clnt_stat err)
{
  busy = false;
  if (err) {
    warn << "err doing sync rpc: " << (int) err << " " << err << "\n";
    return;
  }

  // get new hostlist if it's new
  if (res->resok->hostlistver > hostlist::only().get_version()) {

    // fetch new hostlist from remote host.
  }

  // XXX if the hostlist is changed, we should resync incase the new
  //     nodes have blacklists.
}

void
controlclient::advance_next() 
{
  if (!nexthost) 
    nexthost = hostlist::only().first();
  assert (nexthost);
  
  if (hostlist::only().size() < 2)
    return;
  
  nexthost = hostlist::only().next_circular(nexthost);
  while (!isalive(nexthost->name, nexthost->port) || 
	 myname == nexthost->name && myport == nexthost->port){
    
    nexthost = hostlist::only().next_circular(nexthost);
  }

  assert(nexthost);
}

bool
controlclient::isalive(str name, int port)
{
  struct hostentry *p = hostlist::only().find(name,port);
  if (!p)
    return false;

  ptr<location> loc = chordnode->locations->lookup(p->id);
  if (!loc || !loc->alive())
    return false;
  
  return true;
}
