/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include "stdinc.h"
#include "ns.h"
#include "lrucache.h"
#include "cfs.h"

// the identifier for the ihash class
struct hashID {
  hashID () {}
  hash_t operator() (const bigint &ID) const {
    return ID.getui ();
  }
};

static lrucache<bigint, ptr<cfs_data>, hashID> cfs_cache;

static ptr<cfs_data>
cfs_decode (ptr<ns_block> b)
{
  ptr<cfs_data> data = New refcounted<cfs_data>;
  xdrmem x (b->data, b->len, XDR_DECODE);
  if (xdr_cfs_data (x.xdrp (), data))
    return data;
  else
    return NULL;
}

void
ns_retrieve_cb (cfs_retrieve_t cb, bigint n, bool ro, ptr<ns_block> b)
{
  if (!b)
    cb ((ptr<cfs_data>) 0);
  else {
    ptr<cfs_data> c = cfs_decode (b);
    if (ro && c) {
      cfs_cache.remove (n);
      cfs_cache.insert (n, c);
    }
    cb (c);
  }
}

void
ns_retrieve_cfs (bigint n, cfs_retrieve_t cb)
{
  nsobj->retrieve_block (n, wrap (ns_retrieve_cb, cb, n, false));
}

static void
retrieve_ro_return (ptr<cfs_data> c, cfs_retrieve_t cb)
{
  cb (c);
}

void
ns_retrieve_cfs_readonly (bigint n, cfs_retrieve_t cb)
{
  ptr<cfs_data> *c = cfs_cache [n];
  if (c) {
    delaycb (0, wrap (retrieve_ro_return, *c, cb));
    return;
  }
  nsobj->retrieve_block (n, wrap (ns_retrieve_cb, cb, n, true));
}

