#include "replica_manager.h"
#include "locationtable.h"

chordID 
get_range_min(ptr<vnode> host_node, chordID replicaID, 
	      unsigned min_reps, unsigned max_reps, int max_down) {
  unsigned alive = 0;
  unsigned non_dead = 0;

  ptr<location> rep = host_node->locations->lookup(replicaID);
  ptr<location> loc = host_node->locations->prev_loc_circular(replicaID);
  assert(rep && loc);

  while(loc && loc->id() != rep->id()) {

    if (loc->alive())
      alive++;
    if (loc->down_duration() < max_down) 
      non_dead++;
    if (alive >= min_reps && non_dead >= max_reps)
      break;

    loc = host_node->locations->prev_loc_circular(loc->id());
  }
  return loc->id();
}

// return the alive replicas for <key>.
void
get_succ_replicas (ptr<vnode> host_node, chordID key, unsigned min, 
		   unsigned max, int max_downtime, 
		   vec<ptr<location> > &replicas, 
		   vec<chordID> &range_min, vec<chordID> &range_max, 
		   bool include_master)
{
  ptr<location> master, loc;
  unsigned livehosts = 0, totalhosts = 0;
  timespec ts;
  clock_gettime (CLOCK_REALTIME, &ts);
  chordID range;

  assert (master = host_node->locations->closestsuccloc( key ));
  loc = host_node->locations->next_loc_circular( master->id() );

  if (include_master) {
    replicas.push_back(master);
    range_min.push_back
      (get_range_min(host_node, master->id(), min, max, max_downtime)); 
  }

  while(loc && loc->id() != master->id() ) {
    if ( loc->down_duration() < max_downtime) {
      if ( loc->alive() ) {
	range = get_range_min(host_node, loc->id(), min, max, max_downtime);
	replicas.push_back(loc);
	range_min.push_back(range);
	range_max.push_back(master->id());
	livehosts++;
      }
      totalhosts++;
    }

    if (livehosts >= min-1 && totalhosts >= max-1) 
      break;
    loc = host_node->locations->next_loc_circular(loc->id());
  }

  if (min > host_node->locations->size())
    warn << "WARNING, not enough locations for replication " 
	 << host_node->succs().size() << " < " << min << "\n";
  if (min > livehosts+1)
    warn << "Not enough live replicas!: " << livehosts << " < " << min << "\n";

  return;
}

void
get_pred_replicas(ptr<vnode> host_node, unsigned min, unsigned max, 
		  int max_downtime, vec<ptr<location> > &replicas, 
		  vec<chordID> &range_min, vec<chordID> &range_max) {
  unsigned livehosts = 0, totalhosts = 0;
  ptr<location> me, loc;
  chordID temp_min;
  
  assert (me = host_node->locations->lookup (host_node->my_ID()));
  assert (loc = host_node->locations->prev_loc_circular(me->id()));
  
  while (loc && loc->id() != me->id()) {
    if (loc->down_duration() < max_downtime) {
      if (loc->alive()) {
	replicas.push_back(loc);
	range_max.push_back(loc->id());
	livehosts++;
      }
      totalhosts++;
    }
    assert (loc = host_node->locations->prev_loc_circular(loc->id()));
    temp_min = loc->id();

    if (livehosts >= min-1 && totalhosts >= max-1)
      break;
  }

  // find minimum
  loc = host_node->locations->lookup (temp_min);
  while (loc->down_duration() > max_downtime && loc->id() != host_node->my_ID() )
    loc = host_node->locations->prev_loc_circular(loc->id());

  for(unsigned i=0; i<livehosts; i++) 
    range_min.push_back(loc->id());

  if (min > host_node->locations->size())
    warn << "WARNING, not enough locations for replication " 
	 << host_node->succs().size() << " < " << min << "\n";
  if (min > livehosts+1)
    warn << "Not enough live replicas!: " << livehosts << " < " << min << "\n";
  return;
}

