#ifndef _MERKLE_SERVER_H_
#define _MERKLE_SERVER_H_

#include "merkle_hash.h"
#include "merkle_node.h"
#include "merkle_tree.h"
#include "merkle_syncer.h"
#include "merkle_sync_prot.h"
#include "bigint.h"
#include "qhash.h"
#include <chord.h>

class RPC_delay_args;

typedef callback<void, const rpc_program &, cbdispatch_t>::ref addHandler_t;
typedef callback<void, ptr<location>, bigint>::ref missingfnc2_t;

// One merkle_server runs for each node of the Chord ring.
//  - i.e., one merkle_server per virtual node
//  - and remember, each virtual node has own database 

struct mtree_wrap {
  ihash_entry <mtree_wrap> link;
  chordID node_id;

  ptr<merkle_tree> tree;
  timeval update_time;

  mtree_wrap(chordID _node_id, ptr<merkle_tree> _tree) { 
    node_id = _node_id; 
    update_mtree(_tree);
  }
  void update_mtree(ptr<merkle_tree> _tree) {
    gettimeofday(&update_time, NULL);
    tree = _tree;
  }
};

class merkle_server {
 protected:
  ptr<dhash_blockstore> storage;
  ihash<chordID, mtree_wrap, &mtree_wrap::node_id, 
    &mtree_wrap::link, hashID> mtree_cache;

  ptr<merkle_tree> get_mtree(chordID node_id, chordID min, chordID max, vec<bigint> exclu_proj);
  
 public:
  vnode *host_node;  // XXX bad -- don't directly rely on dhash!
  missingfnc2_t missingfnc;

  void missing (ptr<location> n, bigint key);
  void doRPC (ptr<location> n, RPC_delay_args *args);
  void dispatch (user_args *a);
  merkle_server (ptr<dhash_blockstore> storage, addHandler_t addHandler, 
		 missingfnc2_t missingfnc, vnode *host_node);
};


#endif /* _MERKLE_SERVER_H_ */
