/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include <ctype.h>
#include <errno.h>
#include "rxx.h"
#include "stdinc.h"
#include "projinterface.h"
#include "pastwatch.h"
#include "path.h"
#include "opts.h"

extern ptr<cfsrw> _cfsrw;
extern str _myname;
extern bool const_repository_op (int command);

vec<str> notify_cmds;
vec<str> notify_files;

static int work = 0;
static int error = 0;

static void
return_to_main (cbi cb, int err)
{
  cb (err);
}

static void
done (cbi cb, int err)
{
  if (err)
    error = 1;

  work--;
  if (work == 0)
    cb (error);
}

static void
mail_commit (str arg, cbi cb)
{
  // make sure arg is clean
  for (unsigned i=0; i<arg.len (); i++) {
    if (!isalnum (arg [i]) &&
	arg [i] != ' ' &&
	arg [i] != '.' &&
	arg [i] != '_' &&
	arg [i] != '-' &&
	arg [i] != '@' &&
	arg [i] != '%') {
      warn << "illegal character in " << arg << "\n";
      cb (0);
      return;
    }
  }

  str s = pastwatch_tmpfile ();
  int fd = open (s, O_CREAT | O_TRUNC | O_RDWR, 0600);
  if (fd < 0) {
    warn << "cannot open tmp file for e-mail notification\n";
    cb (0);
    return;
  }

  struct timeval tv;
  gettimeofday (&tv, 0);
  
  str btag = _cfsrw->getreplica ()->branch_tag
    (_cfsrw->getreplica ()->my_branch ());

  strbuf b;
  b << "Project:    " << armor64A (arg_projid) << "\n";
  b << "Changed by: " << _myname << "\n";
  b << "Date:       " << myctime (tv.tv_sec) << "\n";
  b << "Command:    " << cmd2str (arg_command) << "\n";
  b << "Branch Tag: " << btag << "\n\n";

  b << "Modified files:\n";
  for (unsigned i=0; i<notify_cmds.size (); i++) {
    b << "       " << notify_cmds [i] << " "
      << notify_files [i] << "\n";
  }

  b << "\nLog message:\n";
  b << arg_message << "\n\n";
  str m = b;

  write (fd, m.cstr (), m.len ());
  close (fd);

  str subject = "Pastwatch: commit/import files";

  str cmd = "/usr/bin/mail -s \"" << subject << "\" " << arg << " < " << s;

  if (arg_debugmode)
    warn << cmd << "\n";

  if (system (cmd))
    warn << "sending email: " << strerror (errno) << "\n";

  unlink (s);
  cb (0);
}


void
notify (cbi cb)
{
  if ((arg_command != cmd_commit &&
       arg_command != cmd_import &&
       arg_command != cmd_rename &&
       (arg_command != cmd_add || arg_readonly) &&
       arg_command != cmd_checkout) ||
       _cfsrw->getprojinf ()->get_notify () == "") {
    delaycb (0, wrap (return_to_main, cb, 0));
    return;
  }
  str s = _cfsrw->getprojinf ()->get_notify ();
  
  // format of a notify rule
  // <type> <action> <action specific argument>
  //
  // we should probably support the following notify types
  // C - for commit/import/rename/add
  // E - for checkout
  // R - for release
  //
  // right now the only notify action is "mail", which takes an
  // address as an argument.
  //
  // for example:
  // C mail benjie@lcs.mit.edu

  rxx rx ("^([CcEeRr]) ([a-zA-Z]*) ([\\w\\W\\s]*)$");
  if (rx.search (s)) {
    if ((arg_command == cmd_import || arg_command == cmd_commit ||
	 arg_command == cmd_rename || arg_command == cmd_add) &&
	(rx [1] == "C" || rx [1] == "c") &&
	rx [2] == "mail") {
      work++;
      delaycb (0, wrap (mail_commit, rx [3], wrap (done, cb)));
      return;
    }
  }

  delaycb (0, wrap (return_to_main, cb, 0));
}

