/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#ifndef PASTWATCH_PATH_H
#define PASTWATCH_PATH_H

#include "stdinc.h"
#include "pastwatch.h"
#include "opts.h"

inline str
make_path (str dirpath, str file)
{
  str path;
  if (dirpath != "" && dirpath != ".")
    path = dirpath << "/" << file;
  else
    path = file;
  return path;
}

inline void
splitpath (vec<str> &out, str in)
{
  const char *p = in.cstr ();
  const char *e = p + in.len ();
  const char *n;

  for (;;) {
    while (*p == '/')
      p++;
    for (n = p; n < e && *n != '/'; n++)
      ;
    if (n == p)
      return;
    out.push_back (str (p, n - p));
    p = n;
  }
}

inline str
sanitize_path (str path)
{
  if (!path || path.len () == 0)
    return path;

  unsigned end = path.len ()-1;
  // remove trailing /
  while (path[end] == '/' && end > 0)
    end--;
  if (end == 0 && path[end] == '/') // root
    return "/";

  char s [end+1];
  unsigned i,j;
  for (i=0, j=0; i <= end;) {
    s [j] = path [i];
    j++;
    if (path [i] == '/') {
      while (i <= end && path [i] == '/')
        i++;
    }
    else
      i++;
  }

  str x = str (s, j);
  while (x.len () > 2) {
    if (x[0] == '.' && x[1] == '/')
      x = substr (x, 2);
    else
      break;
  }
  return x;
}

inline str
get_parent_dir (str path)
{
  path = sanitize_path (path);

  if (path == "/")
    return "";

  int i = path.len ()-1;
  while (i >= 0) {
    if (path [i] == '/')
      break;
    i--;
  }
  str dir;
  if (i < 0)
    dir = ".";
  else
    dir = substr (path, 0, i);
  return dir;
}

inline str
get_parent_dir (str path, str &name)
{
  path = sanitize_path (path);

  if (path == "/") {
    name = "/";
    return "";
  }

  int i = path.len ()-1;
  while (i >= 0) {
    if (path [i] == '/')
      break;
    i--;
  }
  str dir;
  if (i < 0) {
    dir = ".";
    name = path;
  }
  else {
    dir = substr (path, 0, i);
    name = substr (path, i+1);
  }
  if (name == ".")
    name = "";
  return dir;
}

inline str
get_dir (str path)
{
  struct stat sb;
  if (lstat (path, &sb) == 0 && (sb.st_mode & S_IFDIR))
    return path;
  return get_parent_dir (path);
}

inline str
get_ops_file (str dir)
{
  str s = dir << "/" << PASTWATCH_REF << "/" << "ops";
  return s;
}

inline str
get_supersedes_file (str dir)
{
  str s = dir << "/" << PASTWATCH_REF << "/" << "supersedes";
  return s;
}

inline str
get_info_file (str dir)
{
  str s = dir << "/" << PASTWATCH_REF << "/" << "info";
  return s;
}

inline str
get_db_file (str dir)
{
  str s = dir << "/" << PASTWATCH_REF << "/" << "db";
  return s;
}

extern int get_cfspath (str path, str &cfspath);

#endif

