/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <errno.h>

#include "pastwatch.h"
#include "io.h"
#include "path.h"
#include "pdb.h"

extern void export_one (bigint handle, str path, cbi cb);

static void
rescue_export_reply (bigint root, str path, bigint handle,
                     bool link, cbi cb, int err)
{
  if (err) {
    cb (err);
    return;
  }

  if (link) {
    if (pmgr->exists (path)) {
      warn << path << " in database, already exists in repository?\n";
      warn << "cannot schedule " << path << " for re-link\n";
      cb (-1);
      return;
    }

    str name;
    str dir = get_parent_dir (path, name);
 
    ptr<add_dirop> a = New refcounted<add_dirop> (name, handle);
    int r = new_dirop (dir, a);
    if (r) {
      cb (r);
      return;
    }
  }

  cb (0);
}

static void
rescue_getattr_reply (bigint root, str path, bigint handle, bool link,
                      cbi cb, cfsrcallp cc, cfsrw_status status)
{
  if (status) {
    warn << "cannot get attribute of " << armor64A (handle) << "\n";
    cb (-1);
    return;
  }

  cfsr_getattr_res *res = static_cast<cfsr_getattr_res *> (cc->res ());
  if (res->attr->nlink != 0) {
    warn << armor64A (handle) << " is still linked\n";
    cb (-1);
    return;
  }
  else if (res->attr->type != CFSREG) {
    warn << "cannot rescue directories or symbolic links\n";
    cb (-1);
    return;
  }

  struct stat sb;
  int err = lstat (path, &sb);
  if (err && errno != ENOENT) {
    warn << "cannot stat(7) " << path << "\n";
    cb (-1);
    return;
  }
  if (!err) {
    warn << path << " already exists\n";
    cb (-1);
    return;
  }
  
  export_one (handle, path,
              wrap (rescue_export_reply, root, path, handle, link, cb));
}

void
rescue (bigint root, str path, bigint handle, cbi cb)
{
  cfsrcallp c = New refcounted<cfsrcall>
    (CFSRPROC_GETATTR,
     wrap (rescue_getattr_reply, root, path, handle, arg_relink, cb));
  cfsr_getattr_arg *arg = static_cast<cfsr_getattr_arg *> (c->arg ());
  arg->object = handle;
  cfsr_call (c);
}

