/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#ifndef PASTWATCH_ADB_HH
#define PASTWATCH_ADB_HH

#include "str.h"

#define MAX_LINE_LENGTH 2048

class adb_entry {
public:
  virtual bool match (str) = 0;
  virtual void print_entry (strbuf &b) = 0;
  virtual bool parse_entry (const char *) = 0;
};

class adb_entry_factory {
public:
  virtual ptr<adb_entry> parse_entry (const char *) = 0;
};

class adb {
public:
  str dbname;
  vec<ptr<adb_entry> > entries;
  bool modified;

  adb () { entries.clear (); modified = false; }
  
  int opendb (str fn, adb_entry_factory *ef, bool create = false);
  int closedb ();

  ptr<adb_entry> lookup (str key)
  {
    for (unsigned i=0; i<entries.size (); i++)
      if (entries[i]->match (key))
	return entries[i];
    return 0L;
  }

  int insert (str key, ptr<adb_entry> nelem)
  {
    modified = true;
    for (unsigned i=0; i<entries.size (); i++) {
      if (entries[i]->match (key)) {
	entries[i] = nelem;
	return 0;
      }
    }
    entries.push_back (nelem);
    return 0;
  }

  int del (str key)
  {
    bool found = false;
    vec<ptr<adb_entry> > newvec;

    for (unsigned i=0; i<entries.size (); i++) {
      if (!entries[i]->match (key))
	newvec.push_back (entries[i]);
      else
	found = true;
    }
    if (found) {
      modified = true;
      entries.clear ();
      entries = newvec;
      return 0;
    }
    else
      return -1;
  }
};

#endif

